package com.fossgalaxy.games.tbs.ai.rules;

import com.fossgalaxy.games.tbs.GameState;
import com.fossgalaxy.games.tbs.actions.AbstractAction;
import com.fossgalaxy.games.tbs.entity.Entity;
import com.fossgalaxy.games.tbs.order.Order;
import com.fossgalaxy.games.tbs.parameters.EntityType;
import com.fossgalaxy.games.tbs.ui.GameAction;
import com.fossgalaxy.object.annotations.ObjectDef;
import org.codetome.hexameter.core.api.CubeCoordinate;

import java.util.*;

public class EnsureEntityRule implements ProductionRule {
    private final EntityType producer;
    private final EntityType produced;
    private final int threshold;

    @ObjectDef("EnsureEntityOld")
    public EnsureEntityRule(EntityType producer, EntityType produced, int threshold) {
        this.producer = producer;
        this.produced = produced;
        this.threshold = threshold;
    }

    @ObjectDef("EnsureEntityOld")
    public EnsureEntityRule(EntityType producer, EntityType produced) {
        this(producer, produced, 1);
    }

    @Override
    public Map<UUID, Order> perform(int playerId, GameState state, List<UUID> entities) {

        HashMap<UUID, Order> orders = new HashMap<>();

        Collection<Entity> ourEntities = state.getOwnedEntities(playerId);

        //group entities by type
        Map<EntityType, List<Entity>> entitiesByType = new HashMap<>();
        for (Entity entity : ourEntities) {
            if (entity.getType().isInstance(produced)) {
                entitiesByType.computeIfAbsent(produced, e -> new ArrayList<>()).add(entity);
            }
            if (entity.getType().isInstance(producer)) {
                entitiesByType.computeIfAbsent(producer, e -> new ArrayList<>()).add(entity);
            }
            entitiesByType.computeIfAbsent(entity.getType(), e -> new ArrayList<>()).add(entity);
        }

        //check if we need to build this type...
        int numberOfType = entitiesByType.getOrDefault(produced, Collections.emptyList()).size();
        if (numberOfType >= threshold) {
            return orders;
        }

        //check if we have any workers
        List<Entity> workers = entitiesByType.getOrDefault(producer, Collections.emptyList());
        workers.removeIf(e -> !entities.contains(e.getID()));

        if (workers.isEmpty()) {
            return orders;
        }

        //build the required amount (or less if there aren't enouph workers)
        int numberToBuild = Math.min(threshold - numberOfType, workers.size());
        for (int i = 0; i < numberToBuild; i++) {
            Entity entity = workers.get(i);

            Order buildOrder = findOrderThatBuilds(entity, produced, state);

            if (buildOrder != null) {
                orders.put(entity.getID(), buildOrder);
                buildOrder.doOrder(entity, state);
            }

        }

        return orders;
    }

    public Order findOrderThatBuilds(Entity producer, EntityType produced, GameState state) {

        CubeCoordinate pos = AgentUtils.findBuildingPosition(state, produced, producer.getPos());
        if (pos == null) {
            return null;
        }

        List<GameAction> actions = producer.getType().getAvailableActions();
        for (GameAction action : actions) {
            if (action.getCategory().equals(AbstractAction.CAT_BUILD) && action.appliesTo(produced)) {
                if (action.isPossible(producer, state, pos)) {
                    return action.generateOrder(pos, state);
                }
            }
        }

        return null;

    }

}
